From cd09a2ed8ab222393155141cdf1710bf40b4e0dd Mon Sep 17 00:00:00 2001
From: Shugo Maeda <shugo@ruby-lang.org>
Date: Fri, 29 Oct 2010 19:48:08 +0900
Subject: [PATCH 39/51] supported constants and class variables in refinements.

---
 eval.c              |    2 ++
 include/ruby/ruby.h |    1 +
 insns.def           |    4 ++--
 vm_insnhelper.c     |   20 +++++++++++++++++---
 4 files changed, 22 insertions(+), 5 deletions(-)

diff --git a/eval.c b/eval.c
index 8e0dd97..ede63ff 100644
--- a/eval.c
+++ b/eval.c
@@ -879,9 +879,11 @@ rb_overlay_module(NODE *cref, VALUE klass, VALUE module)
 	    superclass = c;
 	}
     }
+    FL_SET(module, RMODULE_IS_OVERLAYED);
     c = iclass = rb_include_class_new(module, superclass);
     module = RCLASS_SUPER(module);
     while (module) {
+	FL_SET(module, RMODULE_IS_OVERLAYED);
 	c = RCLASS_SUPER(c) = rb_include_class_new(module, RCLASS_SUPER(c));
 	module = RCLASS_SUPER(module);
     }
diff --git a/include/ruby/ruby.h b/include/ruby/ruby.h
index c364476..4bb6068 100644
--- a/include/ruby/ruby.h
+++ b/include/ruby/ruby.h
@@ -630,6 +630,7 @@ struct RClass {
 #define RMODULE_M_TBL(m) RCLASS_M_TBL(m)
 #define RMODULE_SUPER(m) RCLASS_SUPER(m)
 #define RMODULE_HAS_NESTED_METHODS FL_USER2
+#define RMODULE_IS_OVERLAYED FL_USER3
 
 struct RFloat {
     struct RBasic basic;
diff --git a/insns.def b/insns.def
index b77ad7e..3e3b808 100644
--- a/insns.def
+++ b/insns.def
@@ -184,7 +184,7 @@ getclassvariable
 (VALUE val)
 {
     NODE * const cref = rb_vm_get_cref(GET_ISEQ(), GET_LFP(), GET_DFP());
-    val = rb_cvar_get(vm_get_cvar_base(cref), id);
+    val = rb_cvar_get(vm_get_cvar_base(cref, GET_CFP()), id);
 }
 
 /**
@@ -199,7 +199,7 @@ setclassvariable
 ()
 {
     NODE * const cref = rb_vm_get_cref(GET_ISEQ(), GET_LFP(), GET_DFP());
-    rb_cvar_set(vm_get_cvar_base(cref), id, val);
+    rb_cvar_set(vm_get_cvar_base(cref, GET_CFP()), id, val);
 }
 
 /**
diff --git a/vm_insnhelper.c b/vm_insnhelper.c
index accdc91..0b8abe3 100644
--- a/vm_insnhelper.c
+++ b/vm_insnhelper.c
@@ -1172,6 +1172,20 @@ vm_check_if_namespace(VALUE klass)
 }
 
 static inline VALUE
+vm_get_iclass(rb_control_frame_t *cfp, VALUE klass)
+{
+    if (TYPE(klass) == T_MODULE &&
+	FL_TEST(klass, RMODULE_IS_OVERLAYED) &&
+	TYPE(cfp->klass) == T_ICLASS &&
+	RBASIC(cfp->klass)->klass == klass) {
+	return cfp->klass;
+    }
+    else {
+	return klass;
+    }
+}
+
+static inline VALUE
 vm_get_ev_const(rb_thread_t *th, const rb_iseq_t *iseq,
 		VALUE orig_klass, ID id, int is_defined)
 {
@@ -1218,7 +1232,7 @@ vm_get_ev_const(rb_thread_t *th, const rb_iseq_t *iseq,
 
 	/* search self */
 	if (root_cref && !NIL_P(root_cref->nd_clss)) {
-	    klass = root_cref->nd_clss;
+	    klass = vm_get_iclass(th->cfp, root_cref->nd_clss);
 	}
 	else {
 	    klass = CLASS_OF(th->cfp->self);
@@ -1243,7 +1257,7 @@ vm_get_ev_const(rb_thread_t *th, const rb_iseq_t *iseq,
 }
 
 static inline VALUE
-vm_get_cvar_base(NODE *cref)
+vm_get_cvar_base(NODE *cref, rb_control_frame_t *cfp)
 {
     VALUE klass;
 
@@ -1257,7 +1271,7 @@ vm_get_cvar_base(NODE *cref)
 	}
     }
 
-    klass = cref->nd_clss;
+    klass = vm_get_iclass(cfp, cref->nd_clss);
 
     if (NIL_P(klass)) {
 	rb_raise(rb_eTypeError, "no class variables available");
-- 
1.7.0.4

